"""
    item 失败回收中间件
"""
import json
from palp import settings
from palp.pipeline.pipeline import Pipeline


class RedisRecyclePipeline(Pipeline):
    """
        对重试多次失败的 item 进行回收，放入 redis 失败队列
    """

    def pipeline_failed(self, spider, item) -> None:
        """
        回收失败的 item

        :param spider:
        :param item:
        :return:
        """
        from palp.conn import redis_conn

        if redis_conn is None:
            return

        redis_conn.sadd(settings.REDIS_KEY_QUEUE_BAD_ITEM, json.dumps({
            'module': item.__class__.__module__,  # 引用自哪里
            'init': item.__class__.__name__,  # 引用的模块名
            'data': item.to_json()
        }))



"""
    request 去重

    注意：
        setting 启用才会去重
        只有设置 filter_repeat 才会对请求做 阻拦 操作
"""
from palp import settings
from palp.filter.filter_bloom import BloomFilter
from palp.filter.filter_redis_set import RedisSetFilter
from palp.filter.filter_set import SetFilter
from palp.exception import DropRequestException
from palp.filter.filter_redis_bloom import RedisBloomFilter
from palp.middleware.middleware_request import RequestMiddleware


class SetFilterMiddleware(RequestMiddleware):
    """
        基于内存的 set 的去重
    """

    def __init__(self):
        self.request_memory_filter = SetFilter()

    def request_in(self, spider, request):
        if settings.FILTER_REQUEST:
            is_repeat = self.request_memory_filter.is_repeat(obj=request)

            if request.filter_repeat and is_repeat:
                raise DropRequestException(f"丢弃重复请求：{request}")


class BloomFilterMiddleware(RequestMiddleware):
    """
        基于内存的 bloom 的去重
    """

    def __init__(self):
        self.request_bloom_filter = BloomFilter()

    def request_in(self, spider, request):
        if settings.FILTER_REQUEST:
            is_repeat = self.request_bloom_filter.is_repeat(obj=request)

            if request.filter_repeat and is_repeat:
                raise DropRequestException(f"丢弃重复请求：{request}")


class RedisSetFilterMiddleware(RequestMiddleware):
    """
        基于 redis 的 set 的去重
    """

    def __init__(self):
        self.request_redis_filter = RedisSetFilter()

    def request_in(self, spider, request):
        if settings.FILTER_REQUEST:
            is_repeat = self.request_redis_filter.is_repeat(obj=request)

            if request.filter_repeat and is_repeat:
                raise DropRequestException(f"丢弃重复请求：{request}")


class RedisBloomFilterMiddleware(RequestMiddleware):
    """
        基于 redis 的 bloom 去重
    """

    def __init__(self):
        self.request_redis_bloom_filter = RedisBloomFilter()

    def request_in(self, spider, request):
        if settings.FILTER_REQUEST:
            is_repeat = self.request_redis_bloom_filter.is_repeat(obj=request)

            if request.filter_repeat and is_repeat:
                raise DropRequestException(f"丢弃重复请求：{request}")



"""
    单次执行装饰器：执行 request_record、spider_end（分布式时只执行一次）

    执行结果记录：
        本身记录的过程是在中间件 RequestRecordMiddleware 中记录，并更新到 spider_record 变量上
        注意：
            这里侧重的爬虫执行结束之后的结果汇总

        爬虫执行完毕后这要做的事
            本地爬虫：无
            分布式爬虫：发送到 redis 最终多端合并
"""
import json
import datetime
import traceback
from palp import settings
from quickdb import RedisLock
from palp.controller import SpiderController


class SpiderOnceDecorator:
    """
        爬虫单次执行 request_record、spider_end
    """

    @classmethod
    def run_once(cls, spider):
        """
        记录请求量

        :param spider:
        :return:
        """
        from palp.conn import redis_conn

        # 非分布式不上传 redis
        if settings.SPIDER_TYPE == 1:
            cls.run_request_record(spider=spider, record=spider.spider_record)
            cls.run_spider_end(spider=spider)
            return

        # 发送每个 spider 的爬取情况
        redis_conn.hset(settings.REDIS_KEY_RECORD, spider.spider_uuid, json.dumps({
            'all': spider.spider_record['all'],
            'failed': spider.spider_record['failed'],
            'succeed': spider.spider_record['succeed'],
            'stop_time': str(datetime.datetime.now())
        }, ensure_ascii=False))

        # 保证 send_record 只执行一次
        is_end = False
        with RedisLock(conn=redis_conn, lock_name=settings.REDIS_KEY_LOCK + 'Record'):
            # 删除当前 spider 心跳字段
            redis_conn.hdel(settings.REDIS_KEY_HEARTBEAT, spider.spider_uuid)

            # 如果是最后一个心跳，则进行统计记录
            if not redis_conn.exists(settings.REDIS_KEY_HEARTBEAT):
                is_end = True

        # 统计所有机器的结果
        if is_end:
            request_all = 0
            request_failed = 0
            request_succeed = 0

            # 统计
            for spider_uuid, spider_detail in redis_conn.hgetall(settings.REDIS_KEY_RECORD).items():
                spider_detail = json.loads(spider_detail.decode())

                request_all += spider_detail['all']
                request_failed += spider_detail['failed']
                request_succeed += spider_detail['succeed']

            # 给请求中间件发送统计结果
            cls.run_request_record(spider=spider, record={
                'all': request_all,
                'failed': request_failed,
                'succeed': request_succeed,
            })

            # 执行 spider_end
            cls.run_spider_end(spider=spider)

    @staticmethod
    def run_request_record(spider, record: dict):
        """
        调用请求中间件的 request_record 函数发送记录

        record 目前有 3 个字段：all、failed、succeed
        :return:
        """
        for middleware in SpiderController.REQUEST_MIDDLEWARE:
            try:
                middleware.request_record(spider, record)
            except:
                traceback.print_exc()

    @staticmethod
    def run_spider_end(spider):
        """
        执行 spider_end

        :param spider:
        :return:
        """
        for middleware in spider.__class__.SPIDER_MIDDLEWARE:
            try:
                middleware.spider_end(spider)
            except:
                traceback.print_exc()

    def __call__(self, func):
        def _wrapper(*args, **kwargs):
            # 先执行函数
            try:
                func(*args, **kwargs)
            finally:
                # 获取 spider
                spider = args[0]

                # 统计数据
                self.run_once(spider=spider)

        return _wrapper



"""
    spider 检查工具
    用于在请求前进行检查，并进行等待或停止，防止目标网站崩溃等引起的问题

    参数：
        check_time：检查的频率，默认 3s
        failed_limit：检查的容错，默认 3 次，连续 3 次就执行对应函数

    使用方式：
        class Checker(palp.Checker):
            @classmethod
            def need_stop(cls) -> bool:
                return True # 何时需要停止

        class CheckRequestMiddleware(palp.RequestMiddleware):
            def request_in(self, spider, request) -> None:
                Checker.check_time = 3  # 多少 s 检查一次
                Checker.failed_limit = 3    # 连续多少次就执行 wait、stop

                Checker.check_and_wait()    # 当连续错误达到 failed_limit 在之后执行
                Checker.check_and_stop()    # 当连续错误达到 failed_limit 在之后执行（一旦中断无法挽回，只能重启）

    注意：
        分布式时暂时是多个，不能只有一个检查
"""
import time
import threading
from loguru import logger
from abc import abstractmethod
from palp.spider.spider import Spider


class CheckerBase:
    check_time = 3  # 检查的频率，单位 s
    failed_limit = 3  # 容错，连续几次才执行对应的 wait、stop（大于等于 1 的整数）
    _status_cache = []  # 配和容错缓存最近 failed_limit 次状态
    _last_check_time = None  # 上次检查时间
    _stop_status = False  # 停止信号
    _lock = threading.RLock()  # 线程锁
    _has_stop = False  # 是否已经暂停

    @classmethod
    @abstractmethod
    def need_stop(cls) -> bool:
        """
        判断是否需要停止（只需要返回 true、false）

        :return:
        """

    @classmethod
    def _need_stop(cls) -> bool:
        """
        判断是否需要停止（只需要返回 true、false）

        :return:
        """
        if cls.failed_limit < 1:
            cls.failed_limit = 1

        cls._lock.acquire()
        try:
            now_time = time.time()
            if cls._last_check_time is None or now_time - cls._last_check_time > cls.check_time:
                cls._last_check_time = now_time

                # 缓存最近几次状态
                try:
                    status = bool(cls.need_stop())
                except Exception as e:
                    status = True
                    logger.error("need_stop 错误：{}", e)
                cls._status_cache.append(status)

                # 使得缓存数量和容错一致
                if cls._status_cache and len(cls._status_cache) > cls.failed_limit:
                    del cls._status_cache[:len(cls._status_cache) - cls.failed_limit]

                # 根据容错修改状态
                if cls._status_cache.count(True) >= cls.failed_limit:
                    cls._stop_status = True
                else:
                    cls._stop_status = False

        finally:
            cls._lock.release()

        return cls._stop_status

    @classmethod
    def check_and_wait(cls) -> None:
        """
        判断是否需要暂停所有新请求，需要则等待，直到可以继续才释放
        注意：
            当连续错误达到 failed_limit 在之后执行

        :return:
        """
        cls._lock.acquire()

        try:
            while True:
                if cls._need_stop():
                    cls._has_stop = True
                    logger.warning('当前爬虫已暂停')

                    time.sleep(cls.check_time)
                else:
                    if cls._has_stop:
                        logger.warning('当前爬虫已恢复')
                    cls._has_stop = False
                    break
        finally:
            cls._lock.release()

    @classmethod
    def check_and_stop(cls, spider: Spider) -> None:
        """
        判断是否需要停止爬虫执行（暂停所有新请求，并等待已有请求结束）
        注意：
            当连续错误达到 failed_limit 在之后执行
            一旦中断无法挽回，只能重启

        :param spider: spider
        :return:
        """
        cls._lock.acquire()
        try:
            if cls._need_stop() and not cls._has_stop:
                cls._has_stop = True

                for p in spider.spider_controller_list:
                    p.stop = True

                logger.warning('当前爬虫已停止')
        finally:
            cls._lock.release()


class Checker(CheckerBase):

    @classmethod
    @abstractmethod
    def need_stop(cls) -> bool:
        """
        判断是否需要停止（只需要返回 true、false）

        :return:
        """



"""
    通过 redis 实现心跳功能

    master 异常会把自己设置为 slave
    slave 检测到 master 死机会把自己设置为 master
"""
import json
import time
import datetime
import threading
from loguru import logger
from palp import settings
from quickdb import RedisLockNoWait


class ClientHeart:
    def __init__(self, spider):
        """

        :param spider:
        """
        self.spider = spider
        self.beating_time = 4  # 心跳频率
        self.check_time = self.beating_time - 1  # 心跳检查频率

    def start(self):
        beating = threading.Thread(target=self.beating, daemon=True)
        beating.start()

        try:
            self.check_client_beating()
        except Exception as e:
            self.spider.spider_master = False
            logger.exception(e)

    def check_client_beating(self):
        """
        检查各个客户端的心跳，超过 2 次视为停止

        注意：因为不是需要每个都运行，而是同时有一个去运行，所以使用了 RedisLockNoWait
        :return:
        """
        from palp.conn import redis_conn

        while not self.all_client_is_waiting:

            with RedisLockNoWait(conn=redis_conn, lock_name=settings.REDIS_KEY_LOCK + 'CheckHeart',
                                 block_timeout=20) as lock:
                # 判断是否上锁成功
                if not lock.lock_success:
                    continue

                # 获取失败的客户端
                failed_client = [i.decode() for i in redis_conn.smembers(settings.REDIS_KEY_HEARTBEAT_FAILED)]

                # 上锁成功则判断客户端运行情况
                now_time = time.time()
                all_distribute_done = True  # 是否任务分发完毕（只有 master 会分发）
                all_client_is_waiting = True  # 是否所有客户端都无任务处理
                heartbeat = redis_conn.hgetall(settings.REDIS_KEY_HEARTBEAT)
                for client_name, detail in heartbeat.items():
                    client_name = client_name.decode()
                    detail = json.loads(detail.decode())

                    # 校验 2 次失败则为客户端关闭（当前时间-心跳时间-心跳频率 > 心跳频率）
                    if now_time - detail['time'] - self.beating_time > self.beating_time:
                        if client_name in failed_client:
                            logger.warning(f"该客户端异常关闭：{client_name}")
                            redis_conn.srem(settings.REDIS_KEY_HEARTBEAT_FAILED, client_name)
                            redis_conn.hdel(settings.REDIS_KEY_HEARTBEAT, client_name)
                        else:
                            logger.warning(f"该客户端可能异常关闭：{client_name}")
                            redis_conn.sadd(settings.REDIS_KEY_HEARTBEAT_FAILED, client_name)

                            # 检查是否是 master 死机，是的话自己成为 master
                            master_detail = json.loads(redis_conn.get(settings.REDIS_KEY_MASTER).decode())

                            if master_detail and master_detail['name'] == client_name:
                                self.spider.spider_master = True
                                master_detail['name'] = self.spider.spider_uuid
                                redis_conn.set(settings.REDIS_KEY_MASTER, json.dumps(master_detail, ensure_ascii=False))
                    else:
                        if client_name in failed_client:
                            redis_conn.srem(settings.REDIS_KEY_HEARTBEAT_FAILED, client_name)

                        # 判断各项执行状况
                        if detail['waiting'] is False:
                            all_client_is_waiting = False
                        if detail['distribute_done'] is False:
                            all_distribute_done = False
                        logger.debug(f"心跳正常：{client_name}")

                # 如果所有客户端都无任务进行，并且任务分发完毕，则结束
                if heartbeat and all_client_is_waiting and all_distribute_done:
                    logger.debug("所有客户端都已挂起，即将停止")
                    self.stop_all_client()
                    break

                time.sleep(self.check_time)

            time.sleep(0.5)  # 避免访问频繁

    def beating(self):
        """
        保持心跳

        :return:
        """
        from palp.conn import redis_conn

        while True:
            # 检测是否分发完毕
            heart = {
                "time": int(time.time()),
                "waiting": self.spider.all_spider_controller_is_waiting(),
                'distribute_done': self.spider.all_distribute_thread_is_done()
            }

            # 设置 redis 心跳
            redis_conn.hset(
                settings.REDIS_KEY_HEARTBEAT,
                self.spider.spider_uuid,
                json.dumps(heart, ensure_ascii=False)
            )
            time.sleep(self.beating_time)

    @staticmethod
    def stop_all_client():
        """
        发出停止信号，让所有程序停止

        :return:
        """
        from palp.conn import redis_conn

        redis_conn.set(settings.REDIS_KEY_STOP, str(datetime.datetime.now()))

    @property
    def all_client_is_waiting(self) -> bool:
        """
        判断是否所有客户端都陷入了等待

        :return:
        """
        from palp.conn import redis_conn

        return bool(redis_conn.exists(settings.REDIS_KEY_STOP))



"""
    httpx 请求器
"""
import httpx
from loguru import logger
from palp import settings
from palp.network.downloader import ResponseDownloader


class ResponseDownloaderByHttpx(ResponseDownloader):
    """
        适用于 httpx 的请求器
    """
    SESSION = None

    def __new__(cls, *args, **kwargs):
        """
        httpx 设置 session

        注意：
            httpx 的 session 比较麻烦。设置代理必须设置时就加上，建议自定义
            http2 的话也得在这里设置，建议自定义

        :param args:
        :param kwargs:
        """
        if cls.SESSION is None:
            if settings.REQUEST_PROXIES_TUNNEL_URL:
                proxies = {"all://": settings.REQUEST_PROXIES_TUNNEL_URL}
            else:
                proxies = None
            cls.SESSION = httpx.Client(proxies=proxies)

        return object.__new__(cls)

    def response(self):
        """
        处理响应并返回

        :return:
        """
        # 获取请求器
        if self.keep_session:
            request = self.session
        else:
            request = httpx

        # 是否保持 cookie
        if not self.keep_session and self.keep_cookie:
            self.cookies = self.cookie_jar

        # 构建请求参数
        httpx_kwargs = dict(
            url=self.url,
            method=self.method,
            params=self.params,
            data=self.data,
            headers=self.headers,
            cookies=self.cookies,
            timeout=self.timeout,
            json=self.json,
        )

        httpx_kwargs.update(self.kwargs)

        # 请求参数判断
        if self.keep_session and self.proxies:
            logger.warning(
                'httpx 不支持在使用 session 时，使用 proxies 参数，仅支持预设，请自定义：palp.network.downloader_requests.ResponseDownloaderByHttpx'
            )

            del httpx_kwargs['proxies']

        return request.request(**httpx_kwargs)



"""
    单机 spider（不支持分布式）
"""
from palp import settings
from abc import abstractmethod
from palp.spider.spider import Spider
from palp.network.request import Request
from palp.decorator.decorator_spider_wait import SpiderWaitDecorator
from palp.decorator.decorator_spider_once import SpiderOnceDecorator
from palp.decorator.decorator_spider_middleware import SpiderMiddlewareDecorator
from palp.sequence.sequence_memory import PriorityMemorySequence, FIFOMemorySequence


class LocalSpider(Spider):
    """
        单机 spider（不支持分布式）
    """

    def __init__(self, thread_count: int = None, request_filter: bool = False, item_filter: bool = False):
        super().__init__(thread_count, request_filter, item_filter)
        setattr(settings, 'SPIDER_TYPE', 1)

        self.queue = PriorityMemorySequence()  # 请求队列
        self.queue_item = FIFOMemorySequence()  # item 队列
        self.queue_borrow = FIFOMemorySequence()  # 信息传递队列

    @abstractmethod
    def start_requests(self) -> None:
        """
        起始函数

        :return:
        """

    def borrow_request(self, request: Request):
        """
        开启 REQUEST_BORROW 时，用来主动使用资源

        :param request: 新的请求，决定何时调用
        :return:
        """
        # 空 cookie_jar 即视为新的请求，添加 cookie_jar
        if not request.cookie_jar and request.keep_cookie:
            recycle_data = self.queue_borrow.get(block=False)
            if recycle_data:
                if 'cookie_jar' in recycle_data:
                    request.cookie_jar = recycle_data['cookie_jar']

    def recycle_request(self, request: Request):
        """
        开启 REQUEST_BORROW 时，用来主动回收资源，一般在程序执行最后一个函数后调用

        尽量是 request 含有的 key

        :param request: 旧的请求。回收指定资源
        :return:
        """
        recycle_data = {}

        # 回收 cookie
        if request.cookie_jar:
            recycle_data.update({'cookie_jar': request.cookie_jar})

        # 发送回收的资源
        if recycle_data:
            self.queue_borrow.put(recycle_data)

    @SpiderMiddlewareDecorator()
    @SpiderOnceDecorator()
    @SpiderWaitDecorator()
    def run(self) -> None:
        self.start_controller()  # 任务处理
        self.start_distribute()  # 分发任务
        self.wait_distribute_thread_done()  # 等待任务分发完毕



"""
    三种内存队列

    注意：
        python 队列中的 PriorityQueue 需要对比的类实现 __lt__ 方法（小于）不然可能会报错
"""
import queue
from palp import settings
from palp.sequence.sequence import Sequence


class FIFOMemorySequence(Sequence):
    """
        先进先出队列
    """

    def __init__(self, q: queue.Queue = None):
        self.queue = q or queue.Queue()

    def put(self, obj, block=True, timeout=None):
        """
        添加任务

        :param timeout:
        :param block: 为 false 就是 nowait
        :param obj: 数据
        :return:
        """
        self.queue.put(obj, block=block, timeout=timeout)

    def get(self, block=True, timeout=None):
        """
        获取任务

        :param block: 为 false 就是 nowait
        :param timeout:
        :return:
        """
        try:
            item = self.queue.get(block=block, timeout=timeout)
            return item
        except queue.Empty:
            return

    def empty(self):
        """
        判断队列是否为空

        :return:
        """
        return self.queue.empty()

    def qsize(self):
        """
        返回队列大小

        :return:
        """
        return self.queue.qsize()


class LIFOMemorySequence(FIFOMemorySequence):
    """
        后进先出队列
    """

    def __init__(self):
        super().__init__(q=queue.LifoQueue())


class PriorityMemorySequence(FIFOMemorySequence):
    """
        优先级队列
    """

    def __init__(self):
        super().__init__(q=queue.PriorityQueue())

    def put(self, obj, block=True, timeout=None):
        """
        添加任务

        :param block: 为 false 就是 nowait
        :param timeout:
        :param obj: 数据
        :return:
        """
        if hasattr(obj, 'priority'):
            priority = obj.priority
        else:
            priority = settings.DEFAULT_QUEUE_PRIORITY

        self.queue.put([priority, obj], block=block, timeout=timeout)

    def get(self, block=True, timeout=None):
        """
        获取任务

        :param block: 为 false 就是 nowait
        :param timeout:
        :return:
        """
        try:
            item = self.queue.get(block=block, timeout=timeout)
            return item[-1]
        except queue.Empty:
            return



"""
    spider 流程控制器：处理整个爬虫的处理过程流转
"""
import dill
import inspect
import traceback
from loguru import logger
from palp import settings
from threading import Thread
from palp.item.item import Item
from palp.network.request import Request
from palp.network.response import Response
from palp.sequence.sequence import Sequence
from palp.exception import DropRequestException
from palp.tool.short_module import import_module, sort_module


class SpiderController(Thread):
    REQUEST_MIDDLEWARE = []

    def __new__(cls, *args, **kwargs):
        """
        实例化之前加载设置

        :param args:
        :param kwargs:
        """
        if not SpiderController.REQUEST_MIDDLEWARE:
            SpiderController.from_settings()

        return object.__new__(cls)

    def __init__(self, q: Sequence, q_item: Sequence, spider):
        """

        :param q: request 队列
        :param q_item: item 队列
        :param spider: 运行的爬虫
        """
        super().__init__()
        self.queue = q
        self.spider = spider
        self.queue_item = q_item

        self.waiting = False  # 执行状态，如果拿到空任务就是在等待
        self.stop = False  # 是否暂停

    @classmethod
    def from_settings(cls):
        """
        导入设置

        :return:
        """
        modules = sort_module(
            cls_dict=settings.REQUEST_MIDDLEWARE,
            palp_cls_dict=settings.PALP_REQUEST_MIDDLEWARE,
            cls_mapping={
                'REQUEST_FILTER_MIDDLEWARE': settings.REQUEST_FILTER_MIDDLEWARE[settings.SPIDER_TYPE][
                    settings.FILTERING_MODE]
            }
        )

        SpiderController.REQUEST_MIDDLEWARE = import_module(modules)

    def run(self):
        """
        函数执行入口

        :return:
        """
        while not self.stop:
            task = self.queue.get(timeout=1)

            try:
                if task is None:
                    self.waiting = True
                    continue

                self.waiting = False
                self.parse_task(task=task)
            except DropRequestException as e:
                logger.warning(f"丢弃请求：{e.args}")
            except Exception as e:
                if settings.SPIDER_STOP_ON_ERROR:
                    raise

                # spider 报错处理
                for middleware in self.spider.SPIDER_MIDDLEWARE:
                    middleware.spider_error(self.spider, e)

    def parse_task(self, task, request: Request = None, response: Response = None):
        """
        解析处理任务

        :param task: 任务
        :param request: 上一个的请求
        :param response:
        :return:
        """
        # 判断 yield 的类型
        if isinstance(task, Request):
            if request:
                self.add_new_request(task, request, response)
            elif task.jump_spider:
                self.run_jump_spider(task)
            else:
                self.run_requests(task)
        elif issubclass(task.__class__, Item):
            self.queue_item.put(task)
        else:
            logger.warning(f"捕获到非法 yield：{task}")

    def add_new_request(self, new_request: Request, old_request: Request, response: Response):
        """
        添加新的请求到队列中

        :param new_request:
        :param old_request:
        :param response:
        :return:
        """
        # 自动衔接 meta
        if old_request.meta:
            new_request.meta.update(old_request.meta)

        # 自动拼接 url
        if not new_request.url.startswith('http') and response:
            new_request.url = response.urljoin(new_request.url)

        # 添加请求必须参数
        new_request.cookie_jar = old_request.cookie_jar  # 续上上一个的 cookie_jar

        # 修改优先级，深层的函数应该优先处理，避免积压不前（深度爬取）
        if settings.REQUEST_QUEUE_MODE == 3:
            new_request.priority = old_request.priority - 1

        self.queue.put(new_request)

    def run_requests(self, request: Request):
        """
        处理 yield 的 Request

        :param request:
        :return:
        """
        # 判断是否要借 资源
        if settings.REQUEST_BORROW:
            self.spider.borrow_request(request=request)

        # 请求发起前的处理
        for middleware in self.__class__.REQUEST_MIDDLEWARE:
            middleware.request_in(self.spider, request)

        # 处理请求
        response = None
        failed_times = 0
        while failed_times < settings.REQUEST_RETRY_TIMES:
            try:
                response = request.send()
                break
            except Exception as e:
                failed_times += 1

                # 出现错误可直接处理，并返回新的请求，同时阻断当前错误请求传播
                for middleware in self.__class__.REQUEST_MIDDLEWARE:
                    new_request = middleware.request_error(self.spider, request, e)
                    if new_request is None:
                        continue
                    elif isinstance(new_request, Request):
                        self.add_new_request(new_request=new_request, old_request=request, response=response)
                        raise DropRequestException(middleware.__class__.__name__, str(request))
                    else:
                        logger.warning("request_error 仅支持 Request 返回值！")

            # 失败达到最大次数直接丢弃
            if failed_times >= settings.REQUEST_RETRY_TIMES:
                for middleware in self.__class__.REQUEST_MIDDLEWARE:
                    middleware.request_failed(self.spider, request)
                return

        # 请求结束的回调（可在此时判断返回值是否合理，不合理可在此重新构造请求，同时阻断当前请求传播）
        for middleware in self.__class__.REQUEST_MIDDLEWARE:
            new_request = middleware.request_close(self.spider, request, response)
            if new_request is None:
                continue
            elif isinstance(new_request, Request):
                self.add_new_request(new_request=new_request, old_request=request, response=response)
                raise DropRequestException(middleware.__class__.__name__, str(request))
            else:
                logger.warning("request_close 仅支持 Request 返回值！")

        # 继续监控下次 yield
        if request.callback and isinstance(request.callback, str) and hasattr(self.spider, request.callback):
            callback = getattr(self.spider, request.callback)

            if inspect.isgeneratorfunction(callback):  # 很好用的库，判断是否是 yield 函数
                for task in callback(request, response):
                    self.parse_task(task=task, request=request, response=response)
            else:
                callback(request, response)

    def run_jump_spider(self, request: Request):
        """
        处理 jump spider

        :param request:
        :return:
        """
        try:
            # 提取参数
            jump_spider = request.jump_spider
            jump_spider_kwargs = request.jump_spider_kwargs or {}
            jump_request_middleware = request.jump_request_middleware or []

            # 执行 jump_spider
            jump_spider(
                main_spider=self.spider,
                request=request,
                request_middleware=jump_request_middleware,
                **jump_spider_kwargs
            ).run()

            # 执行成功则参数置空，并发起需要的请求
            request.jump_spider = None
            request.jump_spider_kwargs = None
            request.jump_request_middleware = None
            self.queue.put(request)
        except Exception as e:
            traceback.print_exc()

            # 失败直接放入失败中间件，不会重试
            for middleware in self.__class__.REQUEST_MIDDLEWARE:
                middleware.request_failed(self.spider, request)



"""
    配置
        配置优先级 spider > spider setting > palp setting
        三者的配置是覆盖的关系
"""
from pathlib import Path

BASE_PATH = Path(__file__).absolute().parent
SPIDER_STOP_ON_ERROR = True  # spider 在报错时停止（spider、pipeline、middleware）

'''MYSQL'''
# MYSQL_HOST = '127.0.0.1'
# MYSQL_PORT = 3306
# MYSQL_DB = None
# MYSQL_USER = 'root'
# MYSQL_PWD = None
# MYSQL_CONFIG = {}  # 字典形式

'''POSTGRESQL'''
# PG_HOST = '127.0.0.1'
# PG_PORT = 5432
# PG_DB = 'postgres'
# PG_USER = 'postgres'
# PG_PWD = None
# PG_CONFIG = {}  # 字典形式

'''MONGO'''
# MONGO_HOST = '127.0.0.1'
# MONGO_PORT = 27017
# MONGO_USER = None
# MONGO_PWD = None
# MONGO_CONFIG = {}  # 字典形式

'''KAFKA'''
# KAFKA_SERVER = None
# KAFKA_CONFIG = {}  # 字典形式

'''REDIS'''
# REDIS_DB = 0
# REDIS_HOST = '127.0.0.1'
# REDIS_PORT = 6379
# REDIS_PWD = None
# REDIS_CLUSTER_NODES = None  # 如果有集群的情况下，优先集群
# REDIS_MAX_CONNECTIONS = 50  # redis 最大连接数（线程数太大，会连接失败，需要调整）
# REDIS_CONFIG = {}  # redis 连接自定义配置项
# REDIS_POOL_CONFIG = {}  # redis pool 连接自定义配置项

'''请求相关'''
# REQUEST_BORROW = False  # 分发大量任务时，需要在请求中传递的参数（默认回收 cookie 复用）
# REQUEST_BORROW_DELETE_WHEN_START = True  # 启动时删除所有 BORROW（非分布式是在内存里，该选项忽略）
# REQUEST_THREADS = 16  # 线程数量
# REQUEST_FAILED_SAVE = False  # 分布式时保存失败的请求（重试之后仍然失败的）
# REQUEST_RETRY_FAILED = False  # 分布式时启动重试失败请求
# PERSISTENCE_REQUEST_FILTER = False  # 是否持久化请求过滤（分布式时才有效，否则每次结束都会清除）
# REQUEST_DELAY = 0  # 请求间隔，可以是 [0, 3] 代表 0-3s
# REQUEST_RETRY_TIMES = 3  # 请求失败重试次数
# REQUEST_TIMEOUT = 10  # 请求超时时间，也可以是元组 (connect timeout, read timeout)
# RANDOM_USERAGENT = False  # 随机请求头（默认是 computer，指定则开启下面的选项）
# RANDOM_USERAGENT_TYPE = 'computer'  # UA 类型：电脑（computer 代表电脑内随便选，后面代表指定浏览器 chrome、opera、firefox、ie、safari）手机：mobile
# DEFAULT_USER_AGENT = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/107.0.0.0 Safari/537.36 Edg/107.0.1418.35'
# REQUEST_PROXIES_TUNNEL_URL = None  # 隧道代理 url
# RESPONSE_DOWNLOADER = 'palp.network.downloader_requests.ResponseDownloaderByRequests'  # 请求器，这里默认是 requests
# RESPONSE_DOWNLOADER_PARSER = 'palp.network.response_requests.RequestsResponse'  # 解析器，这里默认是 requests
# # RESPONSE_DOWNLOADER = 'palp.network.downloader_httpx.ResponseDownloaderByHttpx'  # 请求器，这里是 httpx
# # RESPONSE_DOWNLOADER_PARSER = 'palp.network.response_httpx.HttpxResponse'  # 解析器，这里是 httpx


'''PIPELINE'''
# ITEM_THREADS = 5  # item 处理的线程数量，默认 5
# ITEM_FAILED_SAVE = False  # 分布式时保存失败的请求（重试之后仍然失败的）
# ITEM_RETRY_FAILED = False  # 分布式时启动重试失败请求
# PIPELINE_ITEM_BUFFER = 0  # 缓存数量，只有当 item 达到一定数量才会入库，0 为不进行缓存
# PIPELINE_RETRY_TIMES = 3  # 入库失败重试次数

# 下载中间件：请求前的处理
PIPELINE = {
    1: "pipelines.pipeline.Pipeline",
}

'''MIDDLEWARE'''
# 请求中间件
REQUEST_MIDDLEWARE = {
    1: "middlewares.middleware.RequestMiddleware",
}

# spider 中间件
SPIDER_MIDDLEWARE = {
    1: 'middlewares.middleware.SpiderMiddleware',
}

'''过滤器'''
# BLOOMFILTER_BIT = 6
# BLOOMFILTER_HASH_NUMBER = 30
# STRICT_FILTER = False  # 严格去重（加锁，会严重影响抓取效率）
# FILTER_ITEM = False  # item 去重
# FILTER_REQUEST = False  # 请求去重
# FILTERING_MODE = 2  # 去重方式：1 为 set 集合，2 为 bloom 过滤（默认）
# PERSISTENCE_ITEM_FILTER = False  # 是否持久化 item 过滤（分布式时才有效，否则每次结束都会清除）

'''队列'''
# REQUEST_QUEUE_MODE = 3  # 1 为先进先出队列，2 为后进先出队列，3 为优先级队列
# DEFAULT_QUEUE_PRIORITY = 300  # 默认的优先级队列的优先级

'''其它'''
# 预警：Email（ palp.send_email ）
# EMAIL_USER = None  # Email 账号
# EMAIL_PWD = None  # Email 授权码

# 预警：DingTalk（ palp.send_dingtalk().xxx ）
# DING_TALK_SECRET = None  # 加签
# DING_TALK_ACCESS_TOKEN = None  # webhook 链接内的 access_token

# 日志（详见：https://loguru.readthedocs.io/en/latest/）
# LOG_LEVEL = "DEBUG"  # 日志过滤等级
# LOG_SHOW = True  # 是否打印日志
# LOG_SAVE = False  # 是否保存日志
LOG_PATH = BASE_PATH.joinpath('log')  # 日志存放地址
# LOG_ROTATION = '500 MB'  # 日志切割方式（日志官网搜索 rotation）
# LOG_RETENTION = '5 days'  # 日志清理时间（日志官网搜索 retention）
# LOG_COMPRESSION = 'zip'  # 日志压缩方式（gz、bz2、xz、lzma、tar、tar.gz、tar.bz2、tar.xz、zip）



"""
    redis item 队列
    使用了 pickle 并使用 zlib 压缩
"""
import zlib
import pickle
from palp import settings
from palp.sequence.sequence import RedisSequence


class FIFOItemRedisSequence(RedisSequence):
    """
        先进先出队列
    """

    @classmethod
    def get_redis_key(cls):
        """
        获取 redis 的键

        :return:
        """
        return settings.REDIS_KEY_QUEUE_ITEM

    def put(self, obj, timeout=None, **kwargs):
        """
        添加任务

        :param obj:
        :param timeout:
        :return:
        """
        from palp.conn import redis_conn

        redis_conn.rpush(self.redis_key, zlib.compress(pickle.dumps(obj)))

    def get(self, timeout=None, **kwargs):
        """
        获取任务（这里是返回的对象）

        :return:
        """
        from palp.conn import redis_conn

        result = redis_conn.blpop(self.redis_key, timeout=timeout)
        if result:
            return pickle.loads(zlib.decompress(result[-1]))

    def empty(self):
        """
        判断队列是否为空

        :return:
        """
        from palp.conn import redis_conn

        return redis_conn.llen(self.redis_key) == 0

    def qsize(self):
        """
        返回队列大小

        :return:
        """
        from palp.conn import redis_conn

        return redis_conn.llen(self.redis_key)



"""
    jump spider（本地执行，用于处理比较频繁的循环事件，完全独立正常的 spider）

    jump spider 内置 spider_start、spider_close
    jump spider 参数 request_middleware 可传请求中间件

    注意：
        jump spider 不走 spider、request 中间件、不走 记录装饰器
        jump spider 不能 yield item yield 的都不会处理
        jump spider 最后会执行 jump_out 合并 cookie、meta，可以自定义
"""
import inspect
from palp import settings
from abc import abstractmethod
from typing import Union, List, Callable
from palp.network.request import Request
from palp.network.response import Response
from palp.tool.short_module import import_module
from palp.exception import NotGeneratorFunctionError
from palp.controller.controller_spider_jump import JumpController


class JumpSpider:
    """
        jump spider（用于处理比较频繁的循环事件）

        注意：不走 spider 中间件，不处理 item
    """

    def __init__(self, main_spider, request: Request, request_middleware: Union[List[Callable], Callable] = None,
                 **kwargs):
        """

        :param request: jump 之后需要执行的请求
        :param request_middleware: 请求中间件
        :param kwargs: 需要被 self.xxx 访问到的参数
        """
        self.main_spider = main_spider
        self.request = request
        self.request_middleware = request_middleware
        self.spider_record = {'all': 0, 'failed': 0, 'succeed': 0}
        self.queue = import_module(settings.REQUEST_QUEUE[1][settings.REQUEST_QUEUE_MODE])[0]  # 请求队列

        # 导入一下自定义设置
        for key, value in kwargs.items():
            setattr(self, key, value)

    def spider_start(self) -> None:
        """
        spider 开始时的操作

        :param spider:
        :return:
        """

    def spider_close(self) -> None:
        """
        spider 开始时的操作

        :param spider:
        :return:
        """

    @abstractmethod
    def start_requests(self) -> None:
        """
        起始函数

        :return:
        """

    def start_distribute(self) -> None:
        """
        刚开始的分发任务

        :return:
        """
        # 校验是否是生成器函数
        if not inspect.isgeneratorfunction(self.start_requests):
            raise NotGeneratorFunctionError("start_requests 必须 yield Request!")

        # 检查起始函数发出请求
        for request in self.start_requests():
            if not isinstance(request, Request):
                raise ValueError("start_requests 仅支持 yield Request")

            # 起始函数无 callback 默认添加
            if request.callback is None:
                request.callback = self.parse

            self.queue.put(request)

    def parse(self, request: Request, response: Response) -> None:
        """
        默认的解析函数（仅 start_requests 默认的）

        @result:
        """

    def run(self) -> None:
        """
        执行 jump spider

        :return:
        """
        self.spider_start()

        try:
            # 分发任务
            self.start_distribute()

            # 执行流程
            JumpController(q=self.queue, request_middleware=self.request_middleware, spider=self).run()
        finally:
            self.spider_close()

    def jump_out(self, request: Request):
        """
        jump spider 执行结束之后，将会调用合并 cookie、meta

        :param request:
        :return:
        """
        self.request.cookie_jar.update(request.cookie_jar)
        self.request.meta.update(request.meta)

    def __setattr__(self, key, value):
        """
        设置属性

        :param key:
        :param value:
        :return:
        """
        self.__dict__[key] = value

    def __getattr__(self, key):
        """
        获取属性

        :param key:
        :return:
        """
        if key not in self.__dict__:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{key}'")

        return self.__dict__[key]



"""
    配置
        配置优先级 spider > spider setting > palp setting
        三者的配置是覆盖的关系
"""
from pathlib import Path

BASE_PATH = Path(__file__).absolute().parent
SPIDER_TYPE = 1  # 爬虫的类型（1 airspider 2 分布式 spider）（非用户设置）
SPIDER_STOP_ON_ERROR = True  # spider 在报错时停止（spider、pipeline、middleware）

'''MYSQL'''
MYSQL_HOST = None
MYSQL_PORT = None
MYSQL_DB = None
MYSQL_USER = None
MYSQL_PWD = None
MYSQL_CONFIG = {}  # 字典形式

'''POSTGRESQL'''
PG_HOST = None
PG_PORT = None
PG_DB = None
PG_USER = None
PG_PWD = None
PG_CONFIG = {}  # 字典形式

'''MONGO'''
MONGO_HOST = None
MONGO_PORT = None
MONGO_USER = None
MONGO_PWD = None
MONGO_CONFIG = {}  # 字典形式

'''KAFKA'''
KAFKA_SERVER = None
KAFKA_CONFIG = {}  # 字典形式

'''REDIS'''
REDIS_DB = 0
REDIS_HOST = None
REDIS_PORT = None
REDIS_PWD = None
REDIS_CLUSTER_NODES = None  # 如果有集群的情况下，优先集群
REDIS_MAX_CONNECTIONS = 50  # redis 最大连接数（线程数太大，会连接失败，需要调整）
REDIS_CONFIG = {}  # redis 连接自定义配置项
REDIS_POOL_CONFIG = {}  # redis pool 连接自定义配置项

# 分布式时的 REDIS KEY
REDIS_KEY_MASTER = '{redis_key}:master'  # redis master（分布式时）(str)
REDIS_KEY_LOCK = '{redis_key}:lock'  # redis 锁(str)
REDIS_KEY_STOP = '{redis_key}:stop'  # 停止所有机器运行（分布式时）(str)
REDIS_KEY_QUEUE_REQUEST = '{redis_key}:request'  # request 队列（list、zset）
REDIS_KEY_QUEUE_REQUEST_BORROW = '{redis_key}:requestBorrow'  # list request 传递参数队列（开启 REQUEST_BORROW）
REDIS_KEY_QUEUE_BAD_REQUEST = '{redis_key}:requestFailed'  # request 失败队列（set）
REDIS_KEY_QUEUE_FILTER_REQUEST = '{redis_key}:filter:request'  # request 过滤队列（set、bloom）
REDIS_KEY_QUEUE_ITEM = '{redis_key}:item'  # item 队列（list）
REDIS_KEY_QUEUE_BAD_ITEM = '{redis_key}:itemFailed'  # item 失败队列（set）
REDIS_KEY_QUEUE_FILTER_ITEM = '{redis_key}:filter:item'  # item 过滤队列（set、bloom）
REDIS_KEY_HEARTBEAT = '{redis_key}:heartbeat'  # 机器的心跳（hash）
REDIS_KEY_HEARTBEAT_FAILED = '{redis_key}:heartbeat_failed'  # 校验失败的机器（set）
REDIS_KEY_RECORD = '{redis_key}:record'  # 记录请求 key (hash)

'''请求相关'''
REQUEST_BORROW = False  # 分发大量任务时，需要在请求中传递的参数（默认回收 cookie 复用）
REQUEST_BORROW_DELETE_WHEN_START = True  # 启动时删除所有 BORROW（非分布式是在内存里，该选项忽略）
REQUEST_THREADS = 16  # 线程数量
REQUEST_FAILED_SAVE = False  # 分布式时保存失败的请求（重试之后仍然失败的）
REQUEST_RETRY_FAILED = False  # 分布式时启动重试失败请求
PERSISTENCE_REQUEST_FILTER = False  # 是否持久化请求过滤（分布式时才有效，否则每次结束都会清除）
REQUEST_DELAY = 0  # 请求间隔，可以是 [0, 3] 代表 0-3s
REQUEST_RETRY_TIMES = 3  # 请求失败重试次数
REQUEST_TIMEOUT = 10  # 请求超时时间，也可以是元组 (connect timeout, read timeout)
RANDOM_USERAGENT = False  # 随机请求头（默认是 computer，指定则开启下面的选项）
RANDOM_USERAGENT_TYPE = 'computer'  # UA 类型：电脑（computer 代表电脑内随便选，后面代表指定浏览器 chrome、opera、firefox、ie、safari）手机：mobile
DEFAULT_USER_AGENT = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/107.0.0.0 Safari/537.36 Edg/107.0.1418.35'
REQUEST_PROXIES_TUNNEL_URL = None  # 隧道代理 url
RESPONSE_DOWNLOADER = 'palp.network.downloader_requests.ResponseDownloaderByRequests'  # 请求器，这里默认是 requests
RESPONSE_DOWNLOADER_PARSER = 'palp.network.response_requests.RequestsResponse'  # 解析器，这里默认是 requests
# RESPONSE_DOWNLOADER = 'palp.network.downloader_httpx.ResponseDownloaderByHttpx'  # 请求器，这里是 httpx
# RESPONSE_DOWNLOADER_PARSER = 'palp.network.response_httpx.HttpxResponse'  # 解析器，这里是 httpx


'''PIPELINE'''
ITEM_THREADS = 5  # item 处理的线程数量，默认 5
ITEM_FAILED_SAVE = False  # 分布式时保存失败的请求（重试之后仍然失败的）
ITEM_RETRY_FAILED = False  # 分布式时启动重试失败请求
PIPELINE_ITEM_BUFFER = 0  # 缓存数量，只有当 item 达到一定数量才会入库，0 为不进行缓存
PIPELINE_RETRY_TIMES = 3  # 入库失败重试次数

# 下载中间件：请求前的处理
PIPELINE = {
    1: "palp.pipeline.pipeline.Pipeline",
}

PALP_PIPELINE = {
    'min': {
        'ITEM_FILTER_PIPELINE': 1
    },
    'max': {
        1: ['ITEM_FAILED_SAVE', 'palp.pipeline.pipeline_recycle.RedisRecyclePipeline']
    }
}

'''MIDDLEWARE'''
# 请求中间件
REQUEST_MIDDLEWARE = {
    1: "palp.middleware.middleware_request.RequestMiddleware",
}

# palp 必须的请求中间件，非用户定义（min、max 代表 REQUEST_MIDDLEWARE 最小最大索引）
PALP_REQUEST_MIDDLEWARE = {
    'min': {
        1: "palp.middleware.middleware_request_check.RequestCheckMiddleware",  # 请求检查
        'REQUEST_FILTER_MIDDLEWARE': 2,  # 请求过滤（反着来代表占用，程序内也要传参，就会转化）
        3: 'palp.middleware.middleware_request_record.RequestRecordMiddleware'  # 记录请求
    },
    'max': {
        1: ['REQUEST_FAILED_SAVE', 'palp.middleware.middleware_request_recycle.RequestRecycleMiddleware']  # 列表代表判断
    }
}

# spider 中间件
SPIDER_MIDDLEWARE = {
    1: 'palp.middleware.middleware_spider.SpiderMiddleware',
}

# palp 必须的请求中间件，非用户定义
PALP_SPIDER_MIDDLEWARE = {
    'min': {
    },
    'max': {
        1: 'palp.middleware.middleware_spider_recycle.SpiderRecycleMiddleware',  # 资源回收
        2: 'palp.middleware.middleware_spider_end.SpiderEndMiddleware'  # 结束语
    }
}

'''过滤器'''
BLOOMFILTER_BIT = 6
BLOOMFILTER_HASH_NUMBER = 30
STRICT_FILTER = False  # 严格去重（加锁，会严重影响抓取效率）
FILTER_ITEM = False  # item 去重
FILTER_REQUEST = False  # 请求去重
FILTERING_MODE = 2  # 去重方式：1 为 set 集合，2 为 bloom 过滤（默认）
PERSISTENCE_ITEM_FILTER = False  # 是否持久化 item 过滤（分布式时才有效，否则每次结束都会清除）

# item 过滤中间件（1、本地，2、云端分布式）
ITEM_FILTER_PIPELINE = {
    1: {
        1: 'palp.pipeline.pipeline_filter.SetFilterPipeline',  # 本地：set 过滤
        2: 'palp.pipeline.pipeline_filter.BloomFilterPipeline',  # 本地：布隆过滤
    },
    2: {
        1: 'palp.pipeline.pipeline_filter.RedisSetFilterPipeline',  # redis：set 过滤
        2: 'palp.pipeline.pipeline_filter.RedisBloomFilterPipeline'  # redis：布隆过滤
    }
}

# 请求过滤中间件（1、本地，2、云端分布式）
REQUEST_FILTER_MIDDLEWARE = {
    1: {
        1: 'palp.middleware.middleware_request_filter.SetFilterMiddleware',  # 本地：set 过滤
        2: 'palp.middleware.middleware_request_filter.BloomFilterMiddleware',  # 本地：布隆过滤
    },
    2: {
        1: 'palp.middleware.middleware_request_filter.RedisSetFilterMiddleware',  # redis：set 过滤
        2: 'palp.middleware.middleware_request_filter.RedisBloomFilterMiddleware'  # redis：布隆过滤
    }
}

'''队列'''
REQUEST_QUEUE_MODE = 3  # 1 为先进先出队列，2 为后进先出队列，3 为优先级队列
ITEM_QUEUE_MODE = 1  # 1 为先进先出队列，2 为后进先出队列（未使用）
DEFAULT_QUEUE_PRIORITY = 300  # 默认的优先级队列的优先级
REQUEST_QUEUE = {
    1: {
        1: 'palp.sequence.sequence_memory.FIFOMemorySequence',  # 本地：先进先出队列
        2: 'palp.sequence.sequence_memory.LIFOMemorySequence',  # 本地：后进先出队列
        3: 'palp.sequence.sequence_memory.PriorityMemorySequence',  # 本地：优先级队列（通过 request 的 level 指定 越小越高）
    },
    2: {
        1: 'palp.sequence.sequence_redis_request.FIFORequestRedisSequence',  # redis：先进先出队列
        2: 'palp.sequence.sequence_redis_request.LIFORequestRedisSequence',  # redis：后进先出队列
        3: 'palp.sequence.sequence_redis_request.PriorityRequestRedisSequence',  # redis：优先级队列（通过 request 的 level 指定）
    }
}
ITEM_QUEUE = {
    1: {
        1: 'palp.sequence.sequence_memory.FIFOMemorySequence',  # 本地：先进先出队列
        2: 'palp.sequence.sequence_memory.LIFOMemorySequence',  # 本地：后进先出队列
    },
    2: {
        1: 'palp.sequence.sequence_redis_item.FIFOItemRedisSequence',  # redis：先进先出队列
        2: 'palp.sequence.sequence_redis_item.LIFOItemRedisSequence',  # redis：后进先出队列
    }
}

'''其它'''
# 预警：Email（ palp.send_email ）
EMAIL_USER = None  # Email 账号
EMAIL_PWD = None  # Email 授权码

# 预警：DingTalk（ palp.send_dingtalk().xxx ）
DING_TALK_SECRET = None  # 加签
DING_TALK_ACCESS_TOKEN = None  # webhook 链接内的 access_token

# 日志（详见：https://loguru.readthedocs.io/en/latest/）
LOG_LEVEL = "DEBUG"  # 日志过滤等级
LOG_SHOW = True  # 是否打印日志
LOG_SAVE = False  # 是否保存日志
LOG_PATH = BASE_PATH.joinpath('log')  # 日志存放地址
LOG_ROTATION = '500 MB'  # 日志切割方式（日志官网搜索 rotation）
LOG_RETENTION = '5 days'  # 日志清理时间（日志官网搜索 retention）
LOG_COMPRESSION = 'zip'  # 日志压缩方式（gz、bz2、xz、lzma、tar、tar.gz、tar.bz2、tar.xz、zip）

# 导入 用户的 spider 设置
try:
    from settings import *
except:
    pass



"""
    返回带有解析器功能的 Response

    注意：
        这里没有强制实现所有方法，但是一定要有 text、url、cookies
"""

import re
import chardet
from parsel import Selector
from bs4 import BeautifulSoup
from abc import abstractmethod
from urllib.parse import urljoin


class Response:
    """
        根据响应体，解析响应
    """

    def __init__(self, resp):
        """

        :param resp: requests 返回的响应
        """
        self.response = resp
        self._selector = None
        self._bs4 = None

    @property
    @abstractmethod
    def text(self) -> str:
        """
        文本方法

        :return:
        """

    @property
    def content(self) -> bytes:
        """
        二进制 content

        :param args:
        :param kwargs:
        :return:
        """

    def json(self, *args, **kwargs) -> dict:
        """
        json 方法

        :return:
        """

    @property
    def status_code(self) -> int:
        """
        状态码

        :return:
        """

    @property
    @abstractmethod
    def cookies(self) -> dict:
        """
        返回 cookies，实在没有返回空字典： {}

        :return:
        """

    @property
    def headers(self) -> dict:
        """
        返回 请求头

        :return:
        """

    @property
    @abstractmethod
    def url(self) -> str:
        """
        响应 url

        :return:
        """

    @property
    def history(self) -> list:
        """
        响应 历史

        :return:
        """

    @property
    def links(self) -> dict:
        """

        :return:
        """

    @property
    def encoding(self):
        """
        解码格式

        :return:
        """
        return chardet.detect(self.content)["encoding"] or 'utf-8'

    @property
    def selector(self) -> Selector:
        """
        延迟解析，否则严重影响性能

        :return:
        """
        if self._selector is None:
            self._selector = Selector(self.text)

        return self._selector

    def bs4(self, features: str = 'lxml') -> BeautifulSoup:
        """

        :param features: 解析器
        :return:
        """
        if self._bs4 is None:
            self._bs4 = BeautifulSoup(markup=self.text, features=features)

        return self._bs4

    def xpath(self, query: str, namespaces: str = None, **kwargs):
        """
        实现 xpath 选择器功能

        :param query: 匹配规则
        :param namespaces: 命名空间
        :param kwargs:
        :return:
        """
        return self.selector.xpath(query=query, namespaces=namespaces, **kwargs)

    def css(self, query: str):
        """
        实现 css 选择器功能

        :param query: 匹配规则
        :return:
        """
        return self.selector.css(query=query)

    def re(self, pattern: str, flags=0) -> list:
        """
        常用 flags：
            re.I：忽略字母大小写
            re.A：匹配相应的 ASCII 字符类别

        :param pattern: 匹配规则
        :param flags: 修饰符
        :return:
        """
        return re.findall(pattern, self.text, flags=flags)

    def re_first(self, pattern: str, flags=0) -> str:
        """
        常用 flags：
            re.I：忽略字母大小写
            re.A：匹配相应的 ASCII 字符类别

        :param pattern: 匹配规则
        :param flags: 修饰符
        :return:
        """
        result = self.re(pattern=pattern, flags=flags)
        if result:
            return result[0]

    def urljoin(self, url: str):
        """
        合并 url

        :return:
        """
        return urljoin(self.response.url, url)

    def __setattr__(self, key, value):
        self.__dict__[key] = value

    def __getattr__(self, item):
        return self.__dict__.get(item)

    def __str__(self):
        return f'<Response [{self.response.status_code}]>'
